(* File: state.mli *)
(* Author: Chad E Brown *)
(* Created: October 2010 *)

open String
open Syntax

val problemfile : string ref

type probitem =
  | ProbDef of string * stp * trm * (string * string) list * float
  | ProbAx of string * string * trm * (string * string) list * float
  | ProbConj of string * trm * (string * string) list * float
val probsig : probitem list ref
val init_probitem : probitem -> unit

val printfof : bool ref

val conjecturename : string ref
val conjecture : (trm * trm) option ref
type proofkind = TSTP | CoqScript | CoqSPfTerm | HOCore | Model | ModelTrue | IsarScript | PfInfo | PfUseful | PfFormdeps
val mkproofterm : proofkind option ref
val mkprooftermp : unit -> bool
val pfusefulout : string option ref
val pfformdepsout : string option ref
val name_base : (string,unit) Hashtbl.t
val name_base_list : string list ref
val name_tp : (string,stp) Hashtbl.t
val name_trm : (string,(trm * stp) * bool ref) Hashtbl.t
val name_trm_list : (string * trm * stp) list ref
val translucent_defns : bool ref
val name_def : (string,trm) Hashtbl.t
val name_def_all : (string,trm) Hashtbl.t
val name_def_prenorm : (string,trm) Hashtbl.t
val name_hyp : (string,trm) Hashtbl.t
val name_hyp_inv : (trm,string * trm) Hashtbl.t
val assumption_lit : (int,trm * trm) Hashtbl.t
val completep : bool ref
    
val get_fresh_name : stp -> string * trm

val initial_branch : trm list ref
val initial_branch_prenorm : trm list ref

val decomposable : string -> bool

val get_timeout_default : float -> float

val st_include_fun : (string -> unit) ref
val st_find_read_thf_fun : (string -> string -> unit) ref

val declare_base_type : string -> unit

val declare_typed_constant : string -> string -> pretrm -> (string * string) list -> unit
    
val declare_definition : string -> string -> pretrm -> (string * string) list -> unit

val declare_thf_logic_formula : string -> string -> pretrm -> (string * string) list -> unit

(*** Code for enumeration of types and terms ***)
val enum_started : bool ref
val enum_of_started : stp -> bool
val enum_of_start : stp -> unit
val new_type_continuation_rtp : stp -> (stp -> int -> unit) -> unit
val new_type_continuation : (stp -> int -> unit) -> unit
val iter_type_continuations_rtp : stp -> stp -> int -> unit
val iter_type_continuations : stp -> int -> unit
val new_term_continuation_rtp : stp -> (stp list * trm * int -> unit) -> unit
val iter_term_continuations_rtp : stp -> stp list -> trm -> int -> unit
val new_usable_type_rtp : stp -> stp -> int -> unit
val usable_types_rtp : stp -> (stp * int) list
val usable_types : unit -> (stp * stp * int) list
val new_usable_head_rtp : stp -> stp list -> trm -> int -> unit
val usable_heads_rtp : stp -> (stp list * trm * int) list

