(* File: mizartypecheckermain.ml *)
(* ported from satallaxmain.ml Aug 2017 *)

open Syntax
open Mizartypeinfo
open State
open Log
open Error

let debug = false
let mizvarcnt = ref 0
let mizwidenlimit = 7

exception InputHelpError of string
exception InputError of string

let itp = Base("$i")
let setoftm = Lam(Ar(itp,Prop),Ap(Choice(itp),Lam(itp,Ap(Forall(itp),Lam(itp,Ap(Ap(Eq(Prop),Ap(Ap(Name("r2_hidden",Ar(itp,Ar(itp,Prop))),DB(0,itp)),DB(1,itp))),Ap(DB(2,Ar(itp,Prop)),DB(0,itp))))))));;

let coqfile = ref ""
let flag_overrides = ref []
let schedule_files : string list ref = ref []
let preparsed = ref false
let starttime = ref 0.0
let endtime = ref 0.0

let printnumaxiomsflag : bool ref = ref false
let selectaxiomsflag : bool ref = ref false

let fclassifier_file : string option ref = ref None
let tclassifier_file : string option ref = ref None
let simple_file : string option ref = ref None

let sine_tolerance_vary = ref true
let sine_generality_threshold_vary = ref true
let sine_tolerance = ref 1.2 (*** can be set by command line --sine_tolerance t where t should >= 1.0; if none is set then slice over different values. ***)
let sine_depth = ref None (*** can be set by command line --sine_depth d where d should be an >= 0.0; if none is set then slice over different depths (more or less) ***)
let sine_generality_threshold = ref 0 (*** can be set by command line --sine_generality_threshold g; where g should be an integer >= 0 ***)

let help_lines =
[ "Usage:"
; "mizarelaborator [optargs] -create <createmizartypingaxiomsfile> <thffile>"
; "mizarelaborator [optargs] -check <existingmizartypingaxiomsfile> <thffile>"
; "mizarelaborator [optargs] -printinfo <existingmizartypingaxiomsfile>"
; "where [optargs] include:"
; "-maxelaborateiters [int]"
; "-elaboratelim [int]"
; "-elaboratetimeout [float]"
; "-acceptcommonmizartypes"
]

let read_proofkind = function
  "tstp" -> TSTP
| "coqscript" -> CoqScript
| "coqspfterm" -> CoqSPfTerm
| "hocore" -> HOCore
| "model" -> Model
| "modeltrue" -> ModelTrue
| "isar" -> IsarScript
| "info" -> PfInfo
| "useful" -> PfUseful
| "formdeps" -> PfFormdeps
| p -> raise (InputHelpError ("Unknown kind of proof " ^ p ^ " for -p"))

let slaveargs = ref []
    
let enslave args = slaveargs := List.rev_append args !slaveargs

let process_short_cmd_line_arg = function
  'v' -> enslave ["-v"]; verbosity := 5
| 'i' -> enslave ["-i"]; completep := false
|  a  -> raise (InputHelpError ("Unknown command line argument " ^ String.make 1 a))

let set_problemfile p =
  if !problemfile = "" then problemfile := p
  else raise (InputHelpError ("Multiple problem files passed: " ^ !problemfile
  ^ " and " ^ p))

let code_status = function
  | (true , Some _) -> 10, "CounterSatisfiable"
  | (true , None  ) -> 15, "Satisfiable"
  | (false, Some _) -> 20, "Theorem"
  | (false, None  ) -> 25, "Unsatisfiable"

let read_thf_file (f:string) (include_fun : string -> unit) =
  let ch = Lexing.from_channel (if (f = "") then stdin else (open_in f)) in
  let old_include_fun = !st_include_fun in
  st_include_fun := include_fun;
(***  List.iter Tptp_config.process_thf (Tptp_parser.tptp_file Tptp_lexer.token ch); ***)
  ignore (Tptp_parser.tptp_file Tptp_lexer.token ch);
  if (!verbosity > 4) then Printf.printf "Finished reading thf file %s\n" f;
  st_include_fun := old_include_fun
      
let rec find_read_thf_file_r odir dir f =
  let ff = (dir ^ "/" ^ f) in
  if (Sys.file_exists ff) then
    read_thf_file ff (find_read_thf_file odir)
  else if (String.length dir > 1) then
    find_read_thf_file_r odir (Filename.dirname dir) f
  else
    raise (FileNotFound f)
and find_read_thf_file dir f =
  let ff = (dir ^ "/" ^ f) in
  if (Sys.file_exists ff) then
    read_thf_file ff (find_read_thf_file dir)
  else
    begin
      try
	let tptpdir = Sys.getenv "TPTP" in
	let tff = (tptpdir ^ "/" ^ f) in
	if (Sys.file_exists tff) then
	  read_thf_file tff (find_read_thf_file dir)
	else
	  find_read_thf_file_r dir dir f
      with
      | Not_found -> find_read_thf_file_r dir dir f
    end;;

st_find_read_thf_fun := find_read_thf_file;;
	
let handle_problem_file () =
  read_thf_file (!problemfile) (find_read_thf_file (Filename.dirname (!problemfile)))

(*** unfinished ***)      
let mizartypechecker_main solv f =
  problemfile := f;
  handle_problem_file ();
  if solv then (*** only try to give an answer to a conjecture if we were asked to ***)
    begin
    end

(*** unfinished ***)      
let mizartypechecker_elaborate f =
  problemfile := f;
  handle_problem_file ();
  ()

let mizartypechecker_extractexamples f =
  problemfile := f;
  handle_problem_file ();
  ()

let number_p x =
  if String.length x = 0 then
    false
  else
    let c = Char.code x.[0] in
    c >= 48 && c < 58

let mizar_numbers_hack f i = (*** just add more types I think numbers should satisfy but are not currently derived; at this point they do seem to be derived, but it's still faster to just assume them ***)
  if i = 0 && number_p f then
    begin
      let io = Ar(itp,Prop) in
      let fi = Name(f,itp) in
      let q p = Ap(Name(p,io),fi) in
      let bl = [q "v7_ordinal1";q "v1_int_1";q "v3_ordinal1"] in
      Hashtbl.add mizar_func_types f ("numberhack",0,[],bl)
    end

let load_mizar_knowledge f =
  declare_base_type "$i";
  let c = open_in_bin f in
  mizar_symbol_order_pred := input_value c;
  mizar_name_tp := input_value c;
  mizar_pattern_list := input_value c;
  typingaxiomslist := [];
  try
    while true do
      let mta : mizartypingaxiom = input_value c in
      typingaxiomslist := mta::!typingaxiomslist;
      match mta with
      | MizarFunctionType(ax,fl,i,pl,bl) ->
	 List.iter (fun f -> Hashtbl.add mizar_func_types f (ax,i,pl,bl); mizar_numbers_hack f i) fl
      | MizarTypeHierarchyStrict(ax,fl,i,(p::pl),bl) ->
	 let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
	 if not (bl2 = []) then
	   begin
	     List.iter (fun f -> Hashtbl.add mizar_typing f (ax,i,p,pl,bl2)) fl
	   end
      | MizarTypeHierarchy(ax,fl,i,(p::pl),bl) ->
	 let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
	 if not (bl2 = []) then
	   begin
	     List.iter (fun f -> Hashtbl.add mizar_typing f (ax,i,p,pl,bl2)) fl
	   end
      | MizarTypingGeneric(ax,i,pl,ml) ->
	  begin
	   let rec db_occur l i a m =
	     match m with
	     | DB(j,_) when j >= l ->
		let k = j - l in
		if k < i then
		  a.(k) <- true
		else
		  raise Exit
	     | Ap(m1,m2) ->
		db_occur l i a m1;
	       db_occur l i a m2
	     | Lam(_,m1) ->
		db_occur (l+1) i a m1
	     | _ -> ()
	   in
	   let mlsimp =
	     List.filter
	       (fun m -> not (List.mem m pl) &&
		 begin
		   match neg_body m with
		   | Some(nm) ->
		       let (h,s) = head_spine nm in
		       (try ignore (List.find (fun n -> match n with DB(_,_) -> false | _ -> true) s); false with Not_found -> true)
		   | None ->
		       let (h,s) = head_spine m in
		       (try ignore (List.find (fun n -> match n with DB(_,_) -> false | _ -> true) s); false with Not_found -> true)
		 end)
	       ml
	   in
	   if not (mlsimp = []) then
	     begin
	       let pl1 = ref pl in
	       let pr1 = ref [] in
	       while not (!pl1 = []) do
		 match !pl1 with
		 | [] -> () (*** impossible ***)
		 | (prim::plr) ->
		    pl1 := plr;
		   let alldboccur =
		     begin
		       let a = Array.create i false in
		       try
			 db_occur 0 i a prim;
			 for j = 0 to i-1 do
			   if not a.(j) then raise Exit
			 done;
			 true
		       with Exit -> false
		     end
		   in
		   if alldboccur then
		     begin
		       match neg_body prim with
		       | Some(nprim) ->
			  begin
			    let (h,s) = head_spine nprim in
			    match h with
			    | Name(f,_) -> Hashtbl.add mizar_typing_gen_forward f (ax,i,prim,(!pr1 @ !pl1),mlsimp)
			    | _ -> ()
			  end
		       | None ->
			  begin
			    let (h,s) = head_spine prim in
			    match h with
			    | Name(f,_) -> Hashtbl.add mizar_typing_gen_forward f (ax,i,prim,(!pr1 @ !pl1),mlsimp)
			    | _ -> ()
			  end
		     end;
		   pr1 := prim::!pr1
	       done
	     end;
	   List.iter
	     (fun concl ->
	       if not (List.mem concl pl) then
		 let alldboccur =
		   begin
		     let a = Array.create i false in
		     try
		       db_occur 0 i a concl;
		       for j = 0 to i-1 do
			 if not a.(j) then raise Exit
		       done;
		       true
		     with Exit -> false
		   end
		 in
		 if alldboccur then
		   begin
		     match neg_body concl with
		     | Some(nconcl) ->
			begin
			  let (mh,ms) = head_spine nconcl in
			  match ms with
			  | [] -> ()
			  | (a::r) ->
			     let (h,s) = head_spine a in
			     match h with
			     | Name(f,_) -> Hashtbl.add mizar_typing_gen_backward f (ax,i,a,pl,concl)
			     | _ -> ()
			end
		     | None ->
			let (mh,ms) = head_spine concl in
			match ms with
			| [] -> ()
			| (a::r) ->
			   let (h,s) = head_spine a in
			   match h with
			   | Name(f,_) -> Hashtbl.add mizar_typing_gen_backward f (ax,i,a,pl,concl)
			   | _ -> ()
		   end)
	     ml
	 end
      | MizarRedefinition(ax,i,pl,f,s,d) ->
	 Hashtbl.add mizar_redef f (ax,i,pl,s,d)
      | _ -> ()
    done
  with End_of_file ->
    close_in c;
    List.iter
      (fun (nname,i,pl,rtp,al,pd,d) ->
	Hashtbl.add mizar_pattern_table nname (i,pl,rtp,al,pd,d))
      !mizar_pattern_list
      
type gentp = Stp of stp | Mtp of trm list

type mizrecinfo = ((string list * bool) list) * ((string * trm) list * (trm * gentp * (string * trm list) list)) list

type mizar_elab_res =
    BadParse
  | MizarTimeout
  | IncompleteArgs
  | CheckingFailure
  | ElaborationSuccess
  | DontKnow

let mer_string mer =
  match mer with
  | BadParse -> "BadParse"
  | MizarTimeout -> "MizarTimeout"
  | IncompleteArgs -> "IncompleteArgs"
  | CheckingFailure -> "CheckingFailure"
  | ElaborationSuccess -> "ElaborationSuccess"
  | DontKnow -> "DontKnow"
